/*
    This file is part of darktable,
    copyright (c) 2009--2010 johannes hanika.

    darktable is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    darktable is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with darktable.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "common/darktable.h"
#include "common/fswatch.h"
#include "common/image.h"
#include "common/image_cache.h"
#include "common/imageio.h"
#include "common/exif.h"
#include "control/control.h"
#include "control/conf.h"
#include "control/jobs.h"
#include <math.h>
#include <sqlite3.h>
#include <string.h>
#include <strings.h>
#include <stdlib.h>
#include <assert.h>

static int64_t dt_image_debug_malloc_size = 0;

static void *
dt_image_debug_malloc(const void *ptr, const size_t size)
{
#ifdef _DEBUG
  assert(ptr == NULL || ptr == (void *)1);
  pthread_mutex_lock(&darktable.db_insert);
  dt_image_debug_malloc_size += size;
  // dt_image_debug_malloc_size ++;
  pthread_mutex_unlock(&darktable.db_insert);
#endif
  return malloc(size);
}

static void
dt_image_debug_free(void *p, size_t size)
{
#ifdef _DEBUG
  if(!p) return;
  pthread_mutex_lock(&darktable.db_insert);
  dt_image_debug_malloc_size -= size;
  // dt_image_debug_malloc_size --;
  pthread_mutex_unlock(&darktable.db_insert);
#endif
  free(p);
}

void dt_image_write_dt_files(dt_image_t *img)
{
  // write .dt file
  if(img->id > 0 && dt_conf_get_bool("write_dt_files"))
  {
    char filename[520];
    dt_image_full_path(img, filename, 512);
    dt_image_path_append_version(img, filename, 512);
    char *c = filename + strlen(filename);
    sprintf(c, ".dt");
    dt_imageio_dt_write(img->id, filename);
    sprintf(c, ".dttags");
    dt_imageio_dttags_write(img->id, filename);
  }
}

int dt_image_is_ldr(dt_image_t *img)
{
  const char *c = img->filename + strlen(img->filename);
  while(*c != '.' && c > img->filename) c--;
  if(!strcasecmp(c, ".jpg") || !strcasecmp(c, ".png") || !strcasecmp(c, ".ppm") || (img->flags & DT_IMAGE_LDR)) return 1;
  else return 0;
}

void dt_image_film_roll(dt_image_t *img, char *pathname, int len)
{
  if(img->film_id == 1)
  {
    snprintf(pathname, len, "%s", _("single images"));
  }
  else
  {
    sqlite3_stmt *stmt;
    sqlite3_prepare_v2(darktable.db, "select folder from film_rolls where id = ?1", -1, &stmt, NULL);
    sqlite3_bind_int(stmt, 1, img->film_id);
    if(sqlite3_step(stmt) == SQLITE_ROW)
    {
      char *f = (char *)sqlite3_column_text(stmt, 0);
      char *c = f + strlen(f);
      while(c > f && *c  != '/') c--;
      if(*c == '/' && c != f) c++;
      snprintf(pathname, len, "%s", c);
    }
    sqlite3_finalize(stmt);
  }
  pathname[len-1] = '\0';
}

void dt_image_full_path(dt_image_t *img, char *pathname, int len)
{
  if(img->film_id == 1)
  {
    snprintf(pathname, len, "%s", img->filename);
  }
  else
  {
    int rc;
    sqlite3_stmt *stmt;
    rc = sqlite3_prepare_v2(darktable.db, "select folder from film_rolls where id = ?1", -1, &stmt, NULL);
    rc = sqlite3_bind_int(stmt, 1, img->film_id);
    if(sqlite3_step(stmt) == SQLITE_ROW)
    {
      snprintf(pathname, len, "%s/%s", (char *)sqlite3_column_text(stmt, 0), img->filename);
    }
    rc = sqlite3_finalize(stmt);
  }
  pathname[len-1] = '\0';
}

void dt_image_export_path(dt_image_t *img, char *pathname, int len)
{
  if(img->film_id == 1)
  {
    snprintf(pathname, len, "%s", img->filename);
  }
  else
  {
    int rc;
    sqlite3_stmt *stmt;
    rc = sqlite3_prepare_v2(darktable.db, "select folder from film_rolls where id = ?1", -1, &stmt, NULL);
    rc = sqlite3_bind_int(stmt, 1, img->film_id);
    if(sqlite3_step(stmt) == SQLITE_ROW)
    {
      snprintf(pathname, len, "%s/darktable_exported/%s", (char *)sqlite3_column_text(stmt, 0), img->filename);
    }
    rc = sqlite3_finalize(stmt);
  }
  dt_image_path_append_version(img, pathname, len);
  pathname[len-1] = '\0';
}

void dt_image_path_append_version(dt_image_t *img, char *pathname, const int len)
{
  // get duplicate suffix
  int version = 0;
  int rc;
  sqlite3_stmt *stmt;
  rc = sqlite3_prepare_v2(darktable.db, "select id from images where filename = ?1 and film_id = ?2", -1, &stmt, NULL);
  rc = sqlite3_bind_text(stmt, 1, img->filename, strlen(img->filename), SQLITE_TRANSIENT);
  rc = sqlite3_bind_int (stmt, 2, img->film_id);
  while (sqlite3_step(stmt) == SQLITE_ROW)
  {
    int id = sqlite3_column_int(stmt, 0);
    if(id == img->id) break;
    version ++;
  }
  rc = sqlite3_finalize(stmt);
  if(version != 0)
  { // add version information:
    char *c = pathname + strlen(pathname);
    while(*c != '.' && c > pathname) c--;
    snprintf(c, pathname + len - c, "_%02d", version);
    char *c2 = img->filename + strlen(img->filename);
    while(*c2 != '.' && c2 > img->filename) c2--;
    snprintf(c+3, pathname + len - c - 3, "%s", c2);
  }
}

void dt_image_print_exif(dt_image_t *img, char *line, int len)
{
  if(img->exif_exposure >= 0.1f)
    snprintf(line, len, "%.1f'' f/%.1f %dmm iso %d", img->exif_exposure, img->exif_aperture, (int)img->exif_focal_length, (int)img->exif_iso);
  else
    snprintf(line, len, "1/%.0f f/%.1f %dmm iso %d", 1.0/img->exif_exposure, img->exif_aperture, (int)img->exif_focal_length, (int)img->exif_iso);
}

dt_image_buffer_t dt_image_get_matching_mip_size(const dt_image_t *img, const int32_t width, const int32_t height, int32_t *w, int32_t *h)
{
  const float scale = fminf(DT_IMAGE_WINDOW_SIZE/(float)(img->width), DT_IMAGE_WINDOW_SIZE/(float)(img->height));
  int32_t wd = MIN(img->width, (int)(scale*img->width)), ht = MIN(img->height, (int)(scale*img->height));
  if(wd & 0xf) wd = (wd & ~0xf) + 0x10;
  if(ht & 0xf) ht = (ht & ~0xf) + 0x10;
  dt_image_buffer_t mip = DT_IMAGE_MIP4;
  const int32_t wd2 = width + width/2;
  const int32_t ht2 = height + height/2;
  while((int)mip > (int)DT_IMAGE_MIP0 && wd > wd2 && ht > ht2)
  {
    mip--;
    if(wd > 32 && ht > 32)
    { // only if it's not vanishing completely :)
      wd >>= 1;
      ht >>= 1;
    }
    else mip = DT_IMAGE_MIP0;
  }
  *w = wd;
  *h = ht;
  return mip;
}

void dt_image_get_exact_mip_size(const dt_image_t *img, dt_image_buffer_t mip, float *w, float *h)
{
  float wd = img->output_width  ? img->output_width  : img->width,
        ht = img->output_height ? img->output_height : img->height;
  if(darktable.develop->image == img)
  {
    int tmpw, tmph;
    dt_dev_get_processed_size(darktable.develop, &tmpw, &tmph);
    wd = tmpw; ht = tmph;
  }
  if(mip == DT_IMAGE_MIPF)
  { // use input width, mipf is before processing
    wd = img->width;
    ht = img->height;
    const float scale = fminf(DT_IMAGE_WINDOW_SIZE/(float)img->width, DT_IMAGE_WINDOW_SIZE/(float)img->height);
    wd *= scale; ht *= scale;
    while((int)mip < (int)DT_IMAGE_MIP4)
    {
      mip++;
      if(wd > 32 && ht > 32)
      { // only if it's not vanishing completely :)
        wd *= .5;
        ht *= .5;
      }
    }
  }
  else if((int)mip < (int)DT_IMAGE_FULL)
  { // full image is full size, rest downscaled by output size
    int mwd, mht;
    dt_image_get_mip_size(img, mip, &mwd, &mht);
    // const int owd = img->output_width  ? img->output_width  : img->width,
    //           oht = img->output_height ? img->output_height : img->height;
    const int owd = (int)wd, oht = (int)ht;
    const float scale = fminf(mwd/(float)owd, mht/(float)oht);
    wd = owd*scale;
    ht = oht*scale;
  }
  *w = wd;
  *h = ht;
}

void dt_image_get_mip_size(const dt_image_t *img, dt_image_buffer_t mip, int32_t *w, int32_t *h)
{
  int32_t wd = img->width, ht = img->height;
  if((int)mip < (int)DT_IMAGE_FULL)
  {
    const float scale = fminf(DT_IMAGE_WINDOW_SIZE/(float)img->width, DT_IMAGE_WINDOW_SIZE/(float)img->height);
    wd *= scale; ht *= scale;
    // make exact mip possible (almost power of two)
    if(wd & 0xf) wd = (wd & ~0xf) + 0x10;
    if(ht & 0xf) ht = (ht & ~0xf) + 0x10;
    while((int)mip < (int)DT_IMAGE_MIP4)
    {
      mip++;
      if(wd > 32 && ht > 32)
      { // only if it's not vanishing completely :)
        wd >>= 1;
        ht >>= 1;
      }
    }
  }
  *w = wd;
  *h = ht;
}

dt_imageio_retval_t dt_image_preview_to_raw(dt_image_t *img)
{
  int p_wd, p_ht;
  float f_wd, f_ht;
  dt_image_get_mip_size(img, DT_IMAGE_MIPF, &p_wd, &p_ht);
  dt_image_get_exact_mip_size(img, DT_IMAGE_MIPF, &f_wd, &f_ht);

  if(dt_image_alloc(img, DT_IMAGE_MIPF)) return DT_IMAGEIO_CACHE_FULL;
  dt_image_check_buffer(img, DT_IMAGE_MIP4, 4*p_wd*p_ht*sizeof(uint8_t));
  dt_image_check_buffer(img, DT_IMAGE_MIPF, 3*p_wd*p_ht*sizeof(float));

  // convert and store in db.
  if(dt_image_is_ldr(img))
    for(int idx=0;idx < p_wd*p_ht; idx++)
      for(int k=0;k<3;k++) img->mipf[3*idx+2-k] = (1./255.)*img->mip[DT_IMAGE_MIP4][4*idx+k];
  else
    dt_imageio_preview_8_to_f(p_wd, p_ht, img->mip[DT_IMAGE_MIP4], img->mipf);
  dt_image_release(img, DT_IMAGE_MIPF, 'w');
  dt_image_release(img, DT_IMAGE_MIPF, 'r');
  // remember this is no real raw data:
  img->flags |= DT_IMAGE_THUMBNAIL;
  return DT_IMAGEIO_OK;
}

dt_imageio_retval_t dt_image_raw_to_preview(dt_image_t *img, const float *raw)
{
  const int raw_wd = img->width;
  const int raw_ht = img->height;
  int p_wd, p_ht;
  float f_wd, f_ht;
  dt_image_get_mip_size(img, DT_IMAGE_MIPF, &p_wd, &p_ht);
  dt_image_get_exact_mip_size(img, DT_IMAGE_MIPF, &f_wd, &f_ht);

  if(dt_image_alloc(img, DT_IMAGE_MIPF)) return DT_IMAGEIO_CACHE_FULL;
  dt_image_check_buffer(img, DT_IMAGE_MIPF, 3*p_wd*p_ht*sizeof(float));

  if(raw_wd == p_wd && raw_ht == p_ht)
  { // use 1:1
    for(int j=0;j<raw_ht;j++) for(int i=0;i<raw_wd;i++)
    {
      const float *cam = raw + 3*(j*raw_wd + i);
      for(int k=0;k<3;k++) img->mipf[3*(j*p_wd + i) + k] = cam[k];
    }
  }
  else
  { // scale to fit
    bzero(img->mipf, 3*p_wd*p_ht*sizeof(float));
    const float scale = fmaxf(raw_wd/f_wd, raw_ht/f_ht);
    for(int j=0;j<p_ht && (int)(scale*j)<raw_ht;j++) for(int i=0;i<p_wd && (int)(scale*i) < raw_wd;i++)
    {
      const float *cam = raw + 3*((int)(scale*j)*raw_wd + (int)(scale*i));
      for(int k=0;k<3;k++) img->mipf[3*(j*p_wd + i) + k] = cam[k];
    }
  }

  dt_image_release(img, DT_IMAGE_MIPF, 'w');
  dt_image_release(img, DT_IMAGE_MIPF, 'r');
  return DT_IMAGEIO_OK;
}

void dt_image_duplicate(const int32_t imgid)
{
  int rc;
  sqlite3_stmt *stmt;
  rc = sqlite3_prepare_v2(darktable.db, "insert into images "
      "(id, film_id, width, height, filename, maker, model, lens, exposure, aperture, iso, "
      "focal_length, datetime_taken, flags, output_width, output_height, crop, "
      "raw_parameters, raw_denoise_threshold, raw_auto_bright_threshold, raw_black, raw_maximum) "
      "select null, film_id, width, height, filename, maker, model, lens, exposure, aperture, iso, "
      "focal_length, datetime_taken, flags, width, height, crop, "
      "raw_parameters, raw_denoise_threshold, raw_auto_bright_threshold, raw_black, raw_maximum "
      "from images where id = ?1", -1, &stmt, NULL);
  rc = sqlite3_bind_int(stmt, 1, imgid);
  rc = sqlite3_step(stmt);
  sqlite3_finalize(stmt);
}

void dt_image_remove(const int32_t imgid)
{
  int rc;
  sqlite3_stmt *stmt;
  rc = sqlite3_prepare_v2(darktable.db, "delete from images where id = ?1", -1, &stmt, NULL);
  rc = sqlite3_bind_int (stmt, 1, imgid);
  rc = sqlite3_step(stmt);
  sqlite3_finalize(stmt);
  rc = sqlite3_prepare_v2(darktable.db, "delete from mipmaps where imgid = ?1", -1, &stmt, NULL);
  rc = sqlite3_bind_int (stmt, 1, imgid);
  rc = sqlite3_step(stmt);
  sqlite3_finalize(stmt);
  rc = sqlite3_prepare_v2(darktable.db, "delete from mipmap_timestamps where imgid = ?1", -1, &stmt, NULL);
  rc = sqlite3_bind_int (stmt, 1, imgid);
  rc = sqlite3_step(stmt);
  sqlite3_finalize(stmt);
  rc = sqlite3_prepare_v2(darktable.db, "update tagxtag set count = count - 1 where "
      "(id2 in (select tagid from tagged_images where imgid = ?2)) or "
      "(id1 in (select tagid from tagged_images where imgid = ?2))", -1, &stmt, NULL);
  rc = sqlite3_bind_int(stmt, 1, imgid);
  rc = sqlite3_step(stmt);
  rc = sqlite3_finalize(stmt);
  rc = sqlite3_prepare_v2(darktable.db, "delete from tagged_images where imgid = ?1", -1, &stmt, NULL);
  rc = sqlite3_bind_int (stmt, 1, imgid);
  rc = sqlite3_step(stmt);
  sqlite3_finalize(stmt);
  rc = sqlite3_prepare_v2(darktable.db, "delete from history where imgid = ?1", -1, &stmt, NULL);
  rc = sqlite3_bind_int (stmt, 1, imgid);
  rc = sqlite3_step(stmt);
  sqlite3_finalize(stmt);
  dt_image_cache_clear(imgid);
}

int dt_image_altered(const dt_image_t *img)
{
  int altered = 0;
  sqlite3_stmt *stmt;
  sqlite3_prepare_v2(darktable.db, "select num from history where imgid = ?1", -1, &stmt, NULL);
  sqlite3_bind_int (stmt, 1, img->id);
  if(sqlite3_step(stmt) == SQLITE_ROW) altered = 1;
  sqlite3_finalize(stmt);
  return altered;
}

int dt_image_import_testlock(dt_image_t *img)
{
  pthread_mutex_lock(&darktable.db_insert);
  int lock = img->import_lock;
  if(!lock) img->import_lock = 1;
  pthread_mutex_unlock(&darktable.db_insert);
  return lock;
}

void dt_image_import_unlock(dt_image_t *img)
{
  pthread_mutex_lock(&darktable.db_insert);
  img->import_lock = 0;
  pthread_mutex_unlock(&darktable.db_insert);
}

int dt_image_reimport(dt_image_t *img, const char *filename, dt_image_buffer_t mip)
{
  if(dt_image_import_testlock(img))
  {
    // fprintf(stderr, "[image_reimport] someone is already loading `%s'!\n", filename);
    return 1;
  }
  if(!img->force_reimport)
  {
    dt_image_buffer_t mip1 = dt_image_get(img, mip, 'r');
    dt_image_release(img, mip1, 'r');
    if(mip1 == mip)
    { // already loaded
      dt_image_import_unlock(img);
      return 0;
    }
  }
  img->output_width = img->output_height = 0;
  dt_imageio_retval_t ret = dt_imageio_open_preview(img, filename);
  if(ret == DT_IMAGEIO_CACHE_FULL)
  { // handle resource conflicts if user provided very small caches:
    dt_image_import_unlock(img);
    return 1;
  }
  else if(ret != DT_IMAGEIO_OK)
  {
    fprintf(stderr, "[image_reimport] could not open %s\n", filename);
    // dt_image_cleanup(img); // still locked buffers. cache will clean itself after a while.
    // dt_image_cache_release(img, 'w');
    dt_image_remove(img->id);
    return 1;
  }

  // fprintf(stderr, "[image_reimport] loading `%s' to fill mip %d!\n", filename, mip);

  // already some db entry there?
  int altered = img->force_reimport;
  img->force_reimport = 0;
  if(dt_image_altered(img)) altered = 1;

  // try loading a .dt[tags] file
  char dtfilename[1031];
  strncpy(dtfilename, filename, 1024);
  dt_image_path_append_version(img, dtfilename, 1024);
  char *c = dtfilename + strlen(dtfilename);
  sprintf(c, ".dttags");
  (void)dt_imageio_dttags_read(img, dtfilename);
  sprintf(c, ".dt");
  if(altered || !dt_imageio_dt_read(img->id, dtfilename))
  {
    dt_develop_t dev;
    dt_dev_init(&dev, 0);
    dt_dev_load_preview(&dev, img);
    dt_dev_process_to_mip(&dev);
    dt_dev_cleanup(&dev);
    // load preview keeps a lock on mipf:
    dt_image_release(img, DT_IMAGE_MIPF, 'r');
  }
  dt_image_import_unlock(img);
  // dt_image_cache_release(imgl, 'w');
  return 0;
}

int dt_image_import(const int32_t film_id, const char *filename)
{
  if(!g_file_test(filename, G_FILE_TEST_IS_REGULAR)) return 0;
  const char *cc = filename + strlen(filename);
  for(;*cc!='.'&&cc>filename;cc--);
  if(!strcmp(cc, ".dt")) return 0;
  if(!strcmp(cc, ".dttags")) return 0;
  char *ext = g_ascii_strdown(cc+1, -1);
  int supported = 0;
  char **extensions = g_strsplit(dt_supported_extensions, ",", 100);
  for(char **i=extensions;*i!=NULL;i++)
    if(!strcmp(ext, *i)) { supported = 1; break; }
  g_strfreev(extensions);
  g_free(ext);
  if(!supported) return 0;
  int rc;
  int ret = 0, id = -1;
  // select from images; if found => return
  gchar *imgfname;
  if(film_id > 1) imgfname = g_path_get_basename((const gchar*)filename);
  else            imgfname = g_build_filename((const gchar*)filename, NULL);
  sqlite3_stmt *stmt;
  rc = sqlite3_prepare_v2(darktable.db, "select id from images where film_id = ?1 and filename = ?2", -1, &stmt, NULL);
  rc = sqlite3_bind_int (stmt, 1, film_id);
  rc = sqlite3_bind_text(stmt, 2, imgfname, strlen(imgfname), SQLITE_STATIC);
  if(sqlite3_step(stmt) == SQLITE_ROW)
  {
    id = sqlite3_column_int(stmt, 0);
    g_free(imgfname);
    rc = sqlite3_finalize(stmt);
    ret = dt_image_open(id); // image already in db, open this.
    if(ret) return 0;
    else return id;
  }
  rc = sqlite3_finalize(stmt);

  // insert dummy image entry in database
  rc = sqlite3_prepare_v2(darktable.db, "insert into images (id, film_id, filename) values (null, ?1, ?2)", -1, &stmt, NULL);
  HANDLE_SQLITE_ERR(rc);
  rc = sqlite3_bind_int (stmt, 1, film_id);
  rc = sqlite3_bind_text(stmt, 2, imgfname, strlen(imgfname), SQLITE_TRANSIENT);
  pthread_mutex_lock(&(darktable.db_insert));
  rc = sqlite3_step(stmt);
  if (rc != SQLITE_DONE) fprintf(stderr, "sqlite3 error %d\n", rc);
  id = sqlite3_last_insert_rowid(darktable.db);
  pthread_mutex_unlock(&(darktable.db_insert));
  rc = sqlite3_finalize(stmt);

  // printf("[image_import] importing `%s' to img id %d\n", imgfname, id);
  dt_image_t *img = dt_image_cache_get_uninited(id, 'w');
  strncpy(img->filename, imgfname, 256);
  img->id = id;
  img->film_id = film_id;

  dt_fswatch_add(darktable.fswatch, DT_FSWATCH_IMAGE, img);
  
  // read dttags and exif for database queries!
  (void) dt_exif_read(img, filename);
  char dtfilename[1024];
  strncpy(dtfilename, filename, 1024);
  dt_image_path_append_version(img, dtfilename, 1024);
  char *c = dtfilename + strlen(dtfilename);
  sprintf(c, ".dttags");
  (void)dt_imageio_dttags_read(img, dtfilename);

  dt_image_cache_flush(img);

  sprintf(c, ".dt");
  (void)dt_imageio_dt_read(img->id, dtfilename);

  g_free(imgfname);

  // single images will probably want to load immediately.
  if(img->film_id == 1) (void)dt_image_reimport(img, filename, DT_IMAGE_MIPF);

  dt_image_cache_release(img, 'w');
  return id;
}

dt_imageio_retval_t dt_image_update_mipmaps(dt_image_t *img)
{
  if(dt_image_lock_if_available(img, DT_IMAGE_MIP4, 'r')) return DT_IMAGEIO_CACHE_FULL;
  int oldwd, oldht;
  dt_image_get_mip_size(img, DT_IMAGE_MIP4, &oldwd, &oldht);
  // create 8-bit mip maps:
  for(dt_image_buffer_t l=DT_IMAGE_MIP3;(int)l>=(int)DT_IMAGE_MIP0;l--)
  {
    int p_wd, p_ht;
    dt_image_get_mip_size(img, l, &p_wd, &p_ht);
    if(dt_image_alloc(img, l)) return DT_IMAGEIO_CACHE_FULL;

    dt_image_check_buffer(img, l, p_wd*p_ht*4*sizeof(uint8_t));
    // printf("creating mipmap %d for img %s: %d x %d\n", l, img->filename, p_wd, p_ht);
    // downscale 8-bit mip
    if(oldwd != p_wd)
      for(int j=0;j<p_ht;j++) for(int i=0;i<p_wd;i++)
        for(int k=0;k<4;k++) img->mip[l][4*(j*p_wd + i) + k] = ((int)img->mip[l+1][8*(2*j)*p_wd + 4*(2*i) + k] + (int)img->mip[l+1][8*(2*j)*p_wd + 4*(2*i+1) + k]
                                                    + (int)img->mip[l+1][8*(2*j+1)*p_wd + 4*(2*i+1) + k] + (int)img->mip[l+1][8*(2*j+1)*p_wd + 4*(2*i) + k])/4;
    else memcpy(img->mip[l], img->mip[l+1], 4*sizeof(uint8_t)*p_ht*p_wd);

    dt_image_release(img, l, 'w');
    dt_image_release(img, l+1, 'r');
  }
  dt_image_release(img, DT_IMAGE_MIP0, 'r');
  return DT_IMAGEIO_OK;
}

void dt_image_init(dt_image_t *img)
{
  // TODO: this is not initialized, in the general case!
  if(strcmp(img->filename,"(unknown)")!=0)  // Reuse of img object...
    dt_fswatch_remove(darktable.fswatch,DT_FSWATCH_IMAGE,img);
  
  for(int k=0;(int)k<(int)DT_IMAGE_MIPF;k++) img->mip[k] = NULL;
  bzero(img->lock, sizeof(dt_image_lock_t)*DT_IMAGE_NONE);
  img->import_lock = 0;
  img->output_width = img->output_height = img->width = img->height = 0;
  img->mipf = NULL;
  img->pixels = NULL;
  img->orientation = -1; // not inited.

  img->black = 0.0f;
  img->maximum = 1.0f;
  img->raw_params.user_flip = -1;
  img->raw_params.med_passes = 0;
  img->raw_params.wb_cam = 0;
  img->raw_params.pre_median = 0;
  img->raw_params.greeneq = 0;
  img->raw_params.no_auto_bright = 0;
  img->raw_params.highlight = 0;
  img->raw_params.demosaic_method = 2;
  img->raw_params.med_passes = 0;
  img->raw_params.four_color_rgb = 0;
  img->raw_params.fill0 = 0;
  img->raw_denoise_threshold = 0.f;
  img->raw_auto_bright_threshold = 0.01f;

  // try to get default raw parameters from db:
  sqlite3_stmt *stmt;
  sqlite3_prepare_v2(darktable.db, "select op_params from presets where operation = 'rawimport' and def=1", -1, &stmt, NULL);
  if(sqlite3_step(stmt) == SQLITE_ROW)
  {
    const void *blob = sqlite3_column_blob(stmt, 0);
    int length = sqlite3_column_bytes(stmt, 0);
    if(length == sizeof(dt_image_raw_parameters_t) + 2*sizeof(float))
      memcpy(&(img->raw_denoise_threshold), blob, length);
  }
  sqlite3_finalize(stmt);
  img->film_id = -1;
  img->flags = 1; // every image has one star. zero is deleted.
  img->id = -1;
  img->cacheline = -1;
  img->force_reimport = 0;
  img->exif_inited = 0;
  bzero(img->exif_maker, sizeof(img->exif_maker));
  bzero(img->exif_model, sizeof(img->exif_model));
  bzero(img->exif_lens, sizeof(img->exif_lens));
  bzero(img->filename, sizeof(img->filename));
  strncpy(img->filename, "(unknown)", 10);
  img->exif_model[0] = img->exif_maker[0] = img->exif_lens[0] = '\0';
  strncpy(img->exif_datetime_taken, "0000:00:00 00:00:00\0", 20);
  img->exif_crop = 1.0;
  img->exif_exposure = img->exif_aperture = img->exif_iso = img->exif_focal_length = 0;
  for(int k=0;(int)k<(int)DT_IMAGE_NONE;k++) img->mip_buf_size[k] = 0;
}

int dt_image_open(const int32_t id)
{
  if(id < 1) return 1;
  dt_image_t *img = dt_image_cache_get(id, 'w');
  if(!img) return 1;
  dt_image_cache_release(img, 'w');
  return 0;
}

int dt_image_open2(dt_image_t *img, const int32_t id)
{ // load stuff from db and store in cache:
  int rc, ret = 1;
  sqlite3_stmt *stmt;
  rc = sqlite3_prepare_v2(darktable.db, "select id, film_id, width, height, filename, maker, model, lens, exposure, aperture, iso, focal_length, datetime_taken, flags, output_width, output_height, crop, raw_parameters, raw_denoise_threshold, raw_auto_bright_threshold, raw_black, raw_maximum from images where id = ?1", -1, &stmt, NULL);
  rc = sqlite3_bind_int (stmt, 1, id);
  // rc = sqlite3_bind_text(stmt, 2, img->filename, strlen(img->filename), SQLITE_STATIC);
  if(sqlite3_step(stmt) == SQLITE_ROW)
  {
    img->id      = sqlite3_column_int(stmt, 0);
    img->film_id = sqlite3_column_int(stmt, 1);
    img->width   = sqlite3_column_int(stmt, 2);
    img->height  = sqlite3_column_int(stmt, 3);
    strncpy(img->filename,   (char *)sqlite3_column_text(stmt, 4), 512);
    strncpy(img->exif_maker, (char *)sqlite3_column_text(stmt, 5), 32);
    strncpy(img->exif_model, (char *)sqlite3_column_text(stmt, 6), 32);
    strncpy(img->exif_lens,  (char *)sqlite3_column_text(stmt, 7), 52);
    img->exif_exposure = sqlite3_column_double(stmt, 8);
    img->exif_aperture = sqlite3_column_double(stmt, 9);
    img->exif_iso = sqlite3_column_double(stmt, 10);
    img->exif_focal_length = sqlite3_column_double(stmt, 11);
    strncpy(img->exif_datetime_taken, (char *)sqlite3_column_text(stmt, 12), 20);
    img->flags = sqlite3_column_int(stmt, 13);
    img->output_width  = sqlite3_column_int(stmt, 14);
    img->output_height = sqlite3_column_int(stmt, 15);
    img->exif_crop = sqlite3_column_double(stmt, 16);
    *(int *)&img->raw_params = sqlite3_column_int(stmt, 17);
    img->raw_denoise_threshold = sqlite3_column_double(stmt, 18);
    img->raw_auto_bright_threshold = sqlite3_column_double(stmt, 19);
    img->black   = sqlite3_column_double(stmt, 20);
    img->maximum = sqlite3_column_double(stmt, 21);
    img->exif_inited = 1;
    
    ret = 0;
  }
  else fprintf(stderr, "[image_open2] failed to open image from database: %s\n", sqlite3_errmsg(darktable.db));
  rc = sqlite3_finalize(stmt);
  if(ret) return ret;
  // add watch for file
  dt_fswatch_add( darktable.fswatch, DT_FSWATCH_IMAGE, img);
  return rc;
}

void dt_image_cleanup(dt_image_t *img)
{
  pthread_mutex_lock(&(darktable.mipmap_cache->mutex));
  for(int k=0;(int)k<(int)DT_IMAGE_NONE;k++) dt_image_free(img, k);
  pthread_mutex_unlock(&(darktable.mipmap_cache->mutex));
}

int dt_image_load(dt_image_t *img, dt_image_buffer_t mip)
{
  if(!img) return 1;
  int ret = 0;
  char filename[1024];
  dt_image_full_path(img, filename, 1024);
  if(mip != DT_IMAGE_FULL &&
    (img->force_reimport || img->width == 0 || img->height == 0))
  {
    ret = dt_image_reimport(img, filename, mip);
  }
  // img not in database. => mip == FULL or kicked out or corrupt or first time load..
  if(mip == DT_IMAGE_MIPF)
  {
    ret = 0;
    if(dt_image_lock_if_available(img, DT_IMAGE_FULL, 'r'))
    {
      if(dt_image_reimport(img, filename, mip)) ret = 1;
      else dt_image_preview_to_raw(img);
    }
    else
    {
      ret = dt_image_raw_to_preview(img, img->pixels);
      dt_image_release(img, DT_IMAGE_FULL, 'r');
    }
  }
  else if(mip == DT_IMAGE_FULL)
  {
    ret = dt_imageio_open(img, filename);
    ret = dt_image_raw_to_preview(img, img->pixels);
    dt_image_release(img, mip, 'r');
  }
  else
  {
    // refuse to load thumbnails for currently developed image.
    if(darktable.develop->image == img) return 1;
    ret = dt_image_reimport(img, filename, mip);
  }
  // TODO: insert abstract hook here?
  dt_control_queue_draw_all();
  return ret;
}

void dt_image_prefetch(dt_image_t *img, dt_image_buffer_t mip)
{
  if(!img || mip > DT_IMAGE_MIPF || mip < DT_IMAGE_MIP0) return;
  pthread_mutex_lock(&(darktable.mipmap_cache->mutex));
  if(img->mip_buf_size[mip] > 0)
  { // already loaded.
    pthread_mutex_unlock(&(darktable.mipmap_cache->mutex));
    return;
  }
  dt_job_t j;
  dt_image_load_job_init(&j, img->id, mip);
  dt_control_revive_job(darktable.control, &j);
  if(!img->lock[mip].write)
  { // start job to load this buf in bg.
    img->lock[mip].write = 1;
    if(dt_control_add_job(darktable.control, &j))
      img->lock[mip].write = 0;
  }
  pthread_mutex_unlock(&(darktable.mipmap_cache->mutex));
}


// =============================
//   mipmap cache functions:
// =============================

void dt_mipmap_cache_init(dt_mipmap_cache_t *cache, int32_t entries)
{
  pthread_mutex_init(&(cache->mutex), NULL);
  for(int k=0;k<(int)DT_IMAGE_NONE;k++)
  {
    cache->total_size[k] = 0;
    int full_entries = dt_conf_get_int ("mipmap_cache_full_images");
    full_entries = MIN(10, MAX(1, full_entries));
    if(k == DT_IMAGE_FULL) entries = full_entries;
    dt_print(DT_DEBUG_CACHE, "[mipmap_cache_init] cache has %d entries for mip %d.\n", entries, k);
    cache->num_entries[k] = entries;
    cache->mip_lru[k] = (dt_image_t **)malloc(sizeof(dt_image_t*)*entries);
    bzero(cache->mip_lru[k], sizeof(dt_image_t*)*entries);
  }
}

void dt_mipmap_cache_cleanup(dt_mipmap_cache_t *cache)
{
  // TODO: free all img bufs?
  for(int k=0;k<(int)DT_IMAGE_NONE;k++) free(cache->mip_lru[k]);
  pthread_mutex_destroy(&(cache->mutex));
}

void dt_mipmap_cache_print(dt_mipmap_cache_t *cache)
{
  int64_t buffers = 0;
  uint64_t bytes = 0;
  for(int k=0;k<(int)DT_IMAGE_NONE;k++)
  {
    int users = 0, write = 0, entries = 0;
    for(int i=0;i<cache->num_entries[k];i++)
    {
      if(cache->mip_lru[k][i])
      {
        // dt_print(DT_DEBUG_CACHE, "[cache entry] buffer %d, image %s locks: %d r %d w\n", k, cache->mip_lru[k][i]->filename, cache->mip_lru[k][i]->lock[k].users, cache->mip_lru[k][i]->lock[k].write);
        entries++;
        users += cache->mip_lru[k][i]->lock[k].users;
        write += cache->mip_lru[k][i]->lock[k].write;
        bytes += cache->mip_lru[k][i]->mip_buf_size[k];
        if(cache->mip_lru[k][i]->mip_buf_size[k]) buffers ++;
      }
    }
    printf("[mipmap_cache] mip %d: fill: %d/%d, users: %d, writers: %d\n", k, entries, cache->num_entries[k], users, write);
    printf("[mipmap_cache] total memory in mip %d: %.2f MB\n", k, cache->total_size[k]/(1024.0*1024.0));
  }
  // printf("[mipmap_cache] occupies %.2f MB in %"PRIi64" (%"PRIi64") buffers\n", bytes/(1024.0*1024.0), buffers, dt_image_debug_malloc_size);
  printf("[mipmap_cache] occupies %.2f MB in %"PRIi64" (%.2f) buffers\n", bytes/(1024.0*1024.0), buffers, dt_image_debug_malloc_size/(1024.0*1024.0));
}

void dt_image_check_buffer(dt_image_t *image, dt_image_buffer_t mip, int32_t size)
{
#ifdef _DEBUG
  assert(image->mip_buf_size[mip] >= size);
#endif
}

int dt_image_alloc(dt_image_t *img, dt_image_buffer_t mip)
{
  int wd, ht;
  dt_image_get_mip_size(img, mip, &wd, &ht);
  size_t size = wd*ht;
  pthread_mutex_lock(&(darktable.mipmap_cache->mutex));
  void *ptr = NULL;
  if     ((int)mip <  (int)DT_IMAGE_MIPF) { size *= 4*sizeof(uint8_t); ptr = (void *)(img->mip[mip]); }
  else if(mip == DT_IMAGE_MIPF) { size *= 3*sizeof(float); ptr = (void *)(img->mipf); }
  else if(mip == DT_IMAGE_FULL) { size *= 3*sizeof(float); ptr = (void *)(img->pixels); }
  else
  {
    pthread_mutex_unlock(&(darktable.mipmap_cache->mutex));
    return 1;
  }
  if(ptr)
  {
    if(size != img->mip_buf_size[mip])
    {
      // can't happen?
      if(0)//img->lock[mip].write || img->lock[mip].users)
      { // in use, can't free this buffer now.
        dt_print(DT_DEBUG_CACHE, "[image_alloc] buffer mip %d of wrong dimensions is still locked! (w:%d u:%d)\n", mip, img->lock[mip].write, img->lock[mip].users);
        pthread_mutex_unlock(&(darktable.mipmap_cache->mutex));
        return 1;
      }
      else
      { // free buffer, alter cache size stats, and continue below.
        dt_image_free(img, mip);
      }
    }
    else
    {
      // dt_print(DT_DEBUG_CACHE, "[image_alloc] locking already allocated image %s\n", img->filename);
      img->lock[mip].write = 1; // write lock
      img->lock[mip].users = 1; // read lock
      pthread_mutex_unlock(&(darktable.mipmap_cache->mutex));
      return 0; // all good, already alloc'ed.
    }
  }

  // printf("allocing %d x %d x %d for %s (%d)\n", wd, ht, size/(wd*ht), img->filename, mip);
  if     ((int)mip <  (int)DT_IMAGE_MIPF) { img->mip[mip] = (uint8_t *)dt_image_debug_malloc(img->mip[mip], size); }
  else if(mip == DT_IMAGE_MIPF) { img->mipf = (float *)dt_image_debug_malloc(img->mipf, size); }
  else if(mip == DT_IMAGE_FULL) { img->pixels = (float *)dt_image_debug_malloc(img->pixels, size); }

  if((mip == DT_IMAGE_FULL && img->pixels == NULL) || (mip == DT_IMAGE_MIPF && img->mipf == NULL) || ((int)mip <  (int)DT_IMAGE_MIPF && img->mip[mip] == NULL))
  {
    fprintf(stderr, "[image_alloc] malloc of %d x %d x %d for image %s mip %d failed!\n", wd, ht, (int)size/(wd*ht), img->filename, mip);
    pthread_mutex_unlock(&(darktable.mipmap_cache->mutex));
    return 1;
  }

  // dt_print(DT_DEBUG_CACHE, "[image_alloc] locking newly allocated image %s of size %f MB\n", img->filename, size/(1024*1024.0));
  
  // garbage collect, free enough space for new buffer:
  dt_mipmap_cache_t *cache = darktable.mipmap_cache;
  // max memory: user supplied number of bytes, evenly distributed among mip levels.
  // clamped between 50MB -- 4GB
#if defined(__APPLE__) || defined(__MACH__)
  // apple's size_t sucks
  size_t max_mem = (size_t)(MAX(52428800, dt_conf_get_int("cache_memory"))/(float)DT_IMAGE_FULL);
#else
  size_t max_mem = (size_t)(MIN(4294967295, MAX(52428800, dt_conf_get_int("cache_memory")))/(float)DT_IMAGE_FULL);
#endif
  dt_print(DT_DEBUG_CACHE, "[image_alloc] mip %d uses %.3f/%.3f MB, alloc %.3f MB\n", mip, 
           cache->total_size[mip]/(1024.0*1024.0), max_mem/(1024.0*1024.0), size/(1024.0*1024.0));
  if(cache->total_size[mip] > 0 && cache->total_size[mip]+size > max_mem)
    for(int k=0;k<cache->num_entries[mip];k++)
  {
    if(cache->mip_lru[mip][k] != NULL &&
      (cache->mip_lru[mip][k]->lock[mip].users == 0 && cache->mip_lru[mip][k]->lock[mip].write == 0))
    {
      dt_image_t *entry = cache->mip_lru[mip][k];
      dt_image_free(entry, mip);
      dt_print(DT_DEBUG_CACHE, "[image_alloc] free mip %d to %.2f MB\n", mip, cache->total_size[mip]/(1024.0*1024.0));
      if(cache->total_size[mip] < 0)
      {
        fprintf(stderr, "[image_alloc] WARNING: memory usage for mip %d dropped below zero!\n", mip);
        cache->total_size[mip] = 0;
      }
      if(cache->total_size[mip] == 0 || cache->total_size[mip]+size < max_mem) break;
    }
  }
  // insert image in node list at newest time
  for(int k=0;k<darktable.mipmap_cache->num_entries[mip];k++)
  {
    if(darktable.mipmap_cache->mip_lru[mip][k] == NULL ||
      (darktable.mipmap_cache->mip_lru[mip][k]->lock[mip].users == 0 && darktable.mipmap_cache->mip_lru[mip][k]->lock[mip].write == 0))
    {
      dt_image_free(darktable.mipmap_cache->mip_lru[mip][k], mip);
      memmove(darktable.mipmap_cache->mip_lru[mip] + k, darktable.mipmap_cache->mip_lru[mip] + k + 1, (darktable.mipmap_cache->num_entries[mip] - k - 1)*sizeof(dt_image_t*));
      darktable.mipmap_cache->mip_lru[mip][darktable.mipmap_cache->num_entries[mip]-1] = img;
      img->lock[mip].write = 1; // write lock
      img->lock[mip].users = 1; // read lock
      img->mip_buf_size[mip] = size;
      cache->total_size[mip] += size;
#if 0
      int allsize = 0;
      for(int k=0;k<darktable.mipmap_cache->num_entries[mip];k++) if(darktable.mipmap_cache->mip_lru[mip][k]) allsize += darktable.mipmap_cache->mip_lru[mip][k]->mip_buf_size[mip];
      printf("[image_alloc] alloc'ed additional %d bytes, now storing %f MB for mip %d\n", size, allsize/(1024.*1024.), mip);
#endif
      pthread_mutex_unlock(&(darktable.mipmap_cache->mutex));
      return 0;
    }
  }
  fprintf(stderr, "[image_alloc] all cache slots seem to be in use! alloc of %d bytes for img id %d mip %d failed!\n", (int)size, img->id, mip);
  for(int k=0;k<darktable.mipmap_cache->num_entries[mip];k++) fprintf(stderr, "[image_alloc] slot[%d] lock %s %d\n", k,darktable.mipmap_cache->mip_lru[mip][k]->lock[mip].write == 0 ? " " : "w", darktable.mipmap_cache->mip_lru[mip][k]->lock[mip].users);
  pthread_mutex_unlock(&(darktable.mipmap_cache->mutex));
  return 1;
}

void dt_image_free(dt_image_t *img, dt_image_buffer_t mip)
{
  // mutex is already locked, as only alloc is allowed to free.
  if(!img) return;
  // dt_image_release(img, mip, "w");
  if((int)mip < (int)DT_IMAGE_MIPF) { if(img->mip[mip] != (uint8_t *)1) dt_image_debug_free(img->mip[mip], img->mip_buf_size[mip]); img->mip[mip] = NULL; }
  else if(mip == DT_IMAGE_MIPF) { if(img->mipf != (float *)1) dt_image_debug_free(img->mipf, img->mip_buf_size[mip]); img->mipf = NULL; }
  else if(mip == DT_IMAGE_FULL) { dt_image_debug_free(img->pixels, img->mip_buf_size[mip]); img->pixels = NULL; }
  else return;
  for(int k=0;k<darktable.mipmap_cache->num_entries[mip];k++)
    if(darktable.mipmap_cache->mip_lru[mip][k] == img) darktable.mipmap_cache->mip_lru[mip][k] = NULL;
  darktable.mipmap_cache->total_size[mip] -= img->mip_buf_size[mip];
  // printf("[image_free] freed %d bytes\n", img->mip_buf_size[mip]);
#ifdef _DEBUG
  if(darktable.control->running)
    assert(img->lock[mip].users == 0);
#endif
  img->mip_buf_size[mip] = 0;
}

int dt_image_lock_if_available(dt_image_t *img, const dt_image_buffer_t mip, const char mode)
{
  if(mip == DT_IMAGE_NONE) return 1;
  pthread_mutex_lock(&(darktable.mipmap_cache->mutex));
  int ret = 0;
  // get image with no write lock set!
  if((int)mip < (int)DT_IMAGE_MIPF)
  {
    if(img->mip[mip] == NULL || img->lock[mip].write) ret = 1;
  }
  else if(mip == DT_IMAGE_MIPF)
  {
    if(img->mipf == NULL || img->lock[mip].write) ret = 1;
  }
  else if(mip == DT_IMAGE_FULL)
  {
    if(img->pixels == NULL || img->lock[mip].write) ret = 1;
  }
  if(ret == 0)
  {
    if(mode == 'w')
    {
      img->lock[mip].write = 1;
      img->lock[mip].users = 1;
    }
    else img->lock[mip].users++;
  }
  pthread_mutex_unlock(&(darktable.mipmap_cache->mutex));
  return ret;
}

dt_image_buffer_t dt_image_get_blocking(dt_image_t *img, const dt_image_buffer_t mip_in, const char mode)
{
  dt_image_buffer_t mip = mip_in;
  if(!img || mip == DT_IMAGE_NONE) return DT_IMAGE_NONE;
  pthread_mutex_lock(&(darktable.mipmap_cache->mutex));
  // get image with no write lock set!
  if((int)mip < (int)DT_IMAGE_MIPF)
  {
    while(mip > 0 && (img->mip[mip] == NULL || img->lock[mip].write)) mip--; // level 0 always there.
  }
  else if(mip == DT_IMAGE_MIPF)
  {
    if(img->mipf == NULL || img->lock[mip].write) mip = DT_IMAGE_NONE;
  }
  else if(mip == DT_IMAGE_FULL)
  {
    if(img->pixels == NULL || img->lock[mip].write) mip = DT_IMAGE_NONE;
  }
  // found?
  if(mip == mip_in)
  {
    if(mode == 'w')
    {
      img->lock[mip].write = 1;
      img->lock[mip].users = 1;
    }
    img->lock[mip].users++;
    pthread_mutex_unlock(&(darktable.mipmap_cache->mutex));
    return mip;
  }
  // already loading? 
  if(img->lock[mip_in].write)
  {
    pthread_mutex_unlock(&(darktable.mipmap_cache->mutex));
    return DT_IMAGE_NONE;
  }
  pthread_mutex_unlock(&(darktable.mipmap_cache->mutex));
 
  // dt_print(DT_DEBUG_CACHE, "[image_get_blocking] requested buffer %d, found %d for image %s locks: %d r %d w\n", mip_in, mip, img->filename, img->lock[mip].users, img->lock[mip].write);

  // start job to load this buf in bg.
  dt_print(DT_DEBUG_CACHE, "[image_get_blocking] reloading mip %d for image %d\n", mip_in, img->id);
  dt_image_load(img, mip_in); // this returns with 'r' locked
  mip = mip_in;

  pthread_mutex_lock(&(darktable.mipmap_cache->mutex));
  if(mip != DT_IMAGE_NONE)
  {
    if(mode == 'w')
    {
      img->lock[mip].write = 1;
      img->lock[mip].users = 1;
    }
    // else img->lock[mip].users++; // already incremented by image_load
  }
  pthread_mutex_unlock(&(darktable.mipmap_cache->mutex));
  return mip;
}

dt_image_buffer_t dt_image_get(dt_image_t *img, const dt_image_buffer_t mip_in, const char mode)
{
  dt_image_buffer_t mip = mip_in;
  if(!img || mip == DT_IMAGE_NONE) return DT_IMAGE_NONE;
  pthread_mutex_lock(&(darktable.mipmap_cache->mutex));
  // get image with no write lock set!
  if((int)mip < (int)DT_IMAGE_MIPF)
  {
    while(mip > 0 && (img->mip[mip] == NULL || img->lock[mip].write)) mip--;
    if(mip == 0 && (img->mip[mip] == NULL || img->lock[mip].write)) mip = DT_IMAGE_NONE;
  }
  else if(mip == DT_IMAGE_MIPF)
  {
    if(img->mipf == NULL || img->lock[mip].write) mip = DT_IMAGE_NONE;
  }
  else if(mip == DT_IMAGE_FULL)
  {
    if(img->pixels == NULL || img->lock[mip].write) mip = DT_IMAGE_NONE;
  }
  if((mip != DT_IMAGE_MIPF && mip != DT_IMAGE_FULL && img->force_reimport) ||
     // (img == darktable.develop->image && darktable.develop->image_force_reload))
     (mip != DT_IMAGE_MIPF && img == darktable.develop->image && darktable.develop->image_force_reload))
        mip = DT_IMAGE_NONE;
  if(mip != DT_IMAGE_NONE)
  {
    if(mode == 'w')
    {
      img->lock[mip].write = 1;
      img->lock[mip].users = 1;
    }
    else img->lock[mip].users++;
  }

  // dt_print(DT_DEBUG_CACHE, "[image_get] requested buffer %d, found %d for image %s locks: %d r %d w\n", mip_in, mip, img->filename, img->lock[mip_in].users, img->lock[mip_in].write);

#if 1
  if(mip != mip_in) // && !img->lock[mip_in].write)
  { // start job to load this buf in bg.
    dt_print(DT_DEBUG_CACHE, "[image_get] reloading mip %d for image %d\n", mip_in, img->id);
    dt_job_t j;
    dt_image_load_job_init(&j, img->id, mip_in);
    // if the job already exists, make it high-priority:
    dt_control_revive_job(darktable.control, &j);
    if(!img->lock[mip_in].write)
    {
      img->lock[mip_in].write = 1;
      if(dt_control_add_job(darktable.control, &j))
        img->lock[mip_in].write = 0;
    }
    if(mip_in == DT_IMAGE_MIP4)
    { // prefetch mipf, don't call prefetch, as we already locked the mutex.
      dt_job_t j;
      dt_image_load_job_init(&j, img->id, DT_IMAGE_MIPF);
      dt_control_revive_job(darktable.control, &j);
      if(!img->lock[DT_IMAGE_MIPF].write)
      { // start job to load this buf in bg.
        img->lock[DT_IMAGE_MIPF].write = 1;
        if(dt_control_add_job(darktable.control, &j))
          img->lock[DT_IMAGE_MIPF].write = 0;
      }
    }
  }
#endif
  pthread_mutex_unlock(&(darktable.mipmap_cache->mutex));
  return mip;
}

void dt_image_release(dt_image_t *img, dt_image_buffer_t mip, const char mode)
{
  pthread_mutex_lock(&(darktable.mipmap_cache->mutex));
  if (mode == 'r' && img->lock[mip].users > 0) img->lock[mip].users --;
  else if (mode == 'w') img->lock[mip].write = 0;  // can only be one writing thread at a time.
  pthread_mutex_unlock(&(darktable.mipmap_cache->mutex));
  // dt_print(DT_DEBUG_CACHE, "[image_release] released lock %c for buffer %d on image %s locks: %d r %d w\n", mode, mip, img->filename, img->lock[mip].users, img->lock[mip].write);
}

